import logging
from typing import Union

from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, YAMLRoot, ElementName, SlotDefinition

from linkml_solr.solrmodel import *

ASSERTED_TYPE_FIELD = '_type'


@dataclass
class Mapper(object):
    """
    Maps between URIs and RDF/SOLR entities and Python datamodel entities
    """
    None

@dataclass
class LinkMLMapper(Mapper):
    """
    LinkML Mapper
    """

    schema: SchemaDefinition

    def _get_slot(self, sn: str) -> Optional[SlotDefinition]:
        for slot in self.schema.slots.values():
            if underscore(slot.name) == sn:
                return slot


    def _get_python_field_for_slot(self, slot: SlotDefinition) -> str:
        return underscore(slot.name) # TODO: map to pythongen

    def pyval_to_solr_atom(self, v: Any, range: ElementName = None, query: SolrQuery = None) -> Union[str, List[str]]:
        if isinstance(v, list):
            return [str(v1) for v1 in v]
        else:
            return str(v)

    def _get_linkml_class(self, in_obj: Dict) -> str:
        if ASSERTED_TYPE_FIELD in in_obj:
            cn = in_obj[ASSERTED_TYPE_FIELD]
            return self.schema.classes[cn]
        else:
            return None

    def _instance_of_linkml_class(self, v) -> bool:
        try:
            type(v).class_name
            return True
        except:
            return False

    def _lookup_slot(self, cls: ClassDefinition, field: str):
        for sn in cls.slots:
            s: SlotDefinition
            s = self.schema.slots[sn]
            if underscore(s.name) == field:
                return s
            if s.alias and underscore(s.alias) == field:
                return s
        logging.error(f'Did not find {field} in {cls.name} slots =  {cls.slots}')

    def _slot_to_solr_prop(self, slot, prefixmap):
        # TODO: allow mapping
        return underscore(slot.name)



from typing import List
import logging
import subprocess
from linkml_runtime.linkml_model.meta import SchemaDefinition, SlotDefinitionName

def _get_multivalued_slots(schema: SchemaDefinition) -> List[SlotDefinitionName]:
    return [s.name for s in schema.slots.values() if s.multivalued]


def bulkload_file(f,
                  format='csv',
                  base_url=None,
                  core=None,
                  schema: SchemaDefinition = None,
                  ):
    """
    Bulkload a file using solr bulkload API

    :param f:
    :param format:
    :param base_url:
    :param core:
    :param schema:
    :return:
    """
    mvslots = _get_multivalued_slots(schema)
    print(f'MV = {mvslots}')
    separator = '%09'
    internal_separator = '%7C'
    parts = [f'f.{s}.split=true&f.{s}.separator={internal_separator}' for s in mvslots]
    url = f'{base_url}/{core}/update?{"&".join(parts)}&commit=true&separator={separator}'
    if format == 'csv':
        ct = 'application/csv'
    elif format == 'json':
        ct = 'application/json'
    else:
        raise Exception(f'Unknown format {format}')
    command = ['curl', url, '-T', f'{f}', '-X', 'POST', '-H', f'Content-type:{ct}']
    print(command)
    subprocess.run(command)




import os
from typing import Union, TextIO, Optional, Dict, Tuple

import click
import json
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, SlotDefinition, ClassDefinitionName
from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.schemaview import SchemaView

from linkml.utils.generator import Generator, shared_arguments

from linkml_solr.solrschema import *

# Map from underlying python data type to solr equivalent
# Note: The underlying types are a union of any built-in python datatype + any type defined in
#       linkml-runtime/utils/metamodelcore.py
# Note the keys are all lower case
solr_schema_types: Dict[str, str] = {
    "int": "int",
    "integer": "int",
    "str": "string",
    "string": "string",
    "bool": "boolean",
    "boolean": "boolean",
    "float": "float",
    "double": "double",
    "decimal": "double",
    "xsddate": "date",
    "xsddatetime": "date",
    "xsdtime": "time"
}

class SolrSchemaGenerator(Generator):
    """
    Generates a Solr schema from a LinkML schema

    The output is a JSON document representing Solr operations such as AddField

    Note that this makes a flattened 'union' schema for all classes, as Solr does not have the concept
    of classes/records
    """
    generatorname = os.path.basename(__file__)
    generatorversion = "0.0.1"
    valid_formats = ["json"]
    visit_all_class_slots = True
    transaction: Transaction = None
    schemaview: SchemaView = None

    def __init__(self, schema: Union[str, TextIO, SchemaDefinition], top_class: Optional[str] = None, **kwargs) -> None:
        super().__init__(schema, **kwargs)
        #self.schema = schema
        self.transaction = Transaction()
        self.post_request = None
        self.inline = False
        self.topCls = top_class  ## SOLR object is one instance of this
        self.entryProperties = {}
        # SOLR-Schema does not have inheritance,
        # so we duplicate slots from inherited parents and mixins
        self.visit_all_slots = True
        self.schemaview = SchemaView(self.schema)

    def _transaction_json(self, transaction: Transaction) -> str:
        obj = json.loads(json_dumper.dumps(transaction))
        # TODO: linkml currently has no way to declare a slot with hyphens in the name
        obj = {k.replace('_', '-'): v for k, v in obj.items() if k != '@type'}
        self.post_request = obj
        return json.dumps(obj, indent='  ')

    def class_schema(self, class_name: Union[str, ClassDefinitionName]) -> str:
        transaction = self.get_transaction(class_name)
        return self._transaction_json(transaction)

    def end_schema(self, **_) -> None:
        transaction = self.get_transaction()
        print(self._transaction_json(transaction))

    def get_transaction(self, class_name: Union[str, ClassDefinitionName] = None) -> Transaction:
        transaction = Transaction()
        sv = self.schemaview
        field_dict = {}
        for cn, c in sv.all_classes().items():
            if class_name is None or str(cn) == str(class_name):
                for s in sv.class_induced_slots(cn):
                    field = self.get_field(s)
                    field_dict[field.name] = field
        for field in field_dict.values():
            transaction.add_field.append(field)
        return transaction

    def get_field(self, slot: SlotDefinition) -> Field:
        range = slot.range
        if range in self.schema.classes:
            range = 'string' # FK
        elif range in self.schema.enums:
            range = 'string'
        elif range in self.schema.types:
            t = self.schema.types[range]
            range = t.typeof
            if t.typeof is None:
                range = t.base
        range = str(range).lower()
        if range in solr_schema_types:
            field = Field(name=slot.name,
                          multiValued=slot.multivalued,
                          type=solr_schema_types[range])
            return field
        else:
            raise Exception(f'Unknown range: {range} for slot {slot}')


@shared_arguments(SolrSchemaGenerator)
@click.command()
def cli(yamlfile, **kwargs):
    """ Generate SOLR Schema representation of a LinkML model """
    print(SolrSchemaGenerator(yamlfile, **kwargs).serialize(**kwargs))


if __name__ == '__main__':
    cli()



import logging
from json import JSONDecodeError
from typing import Union, Dict, Tuple, Type, List
import pysolr
from dataclasses import dataclass
import json
import requests
from linkml_dataops.query.query_model import AbstractQuery, FetchById
from linkml_dataops.query.queryengine import QueryEngine

from linkml_runtime.dumpers import json_dumper
from linkml_runtime.utils.formatutils import underscore
from linkml_runtime.linkml_model.meta import SchemaDefinition, ClassDefinition, YAMLRoot, ElementName, SlotDefinition, SlotDefinitionName

from linkml_solr.solrmodel import SolrEndpoint, SolrQuery, SolrQueryResult, RawSolrResult, FIELD
from linkml_solr.solrschemagen import SolrSchemaGenerator
from linkml_solr.mapper import LinkMLMapper

# https://stackoverflow.com/questions/1176136/convert-string-to-python-class-object
def class_for_name(module_name, class_name):
    # load the module, will raise ImportError if module cannot be loaded
    m = __import__(module_name, globals(), locals(), class_name)
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


@dataclass
class SolrQueryEngine(QueryEngine):
    """
    ORM wrapper for SOLR endpoint
    """

    endpoint: SolrEndpoint = None
    schema: SchemaDefinition = None
    mapper: LinkMLMapper = None
    discriminator_field: SlotDefinitionName = None
    python_classes: List[Type[YAMLRoot]] = None

    def __post_init__(self):
        # TODO: use schemaview
        if self.mapper is None:
            self.mapper = LinkMLMapper(schema=self.schema)
        if self.mapper.schema is None:
            self.mapper.schema = self.schema

    def query(self, target_class: Type[YAMLRoot] = None, **params) -> List[YAMLRoot]:
        """
        As search, but just returns items, discarding facet info etc

        :param target_class:
        :param params:
        :return:
        """
        return self.search(target_class, **params).items

    def search(self, target_class: Type[YAMLRoot] = None,
               solr_params: Dict = None,
               search_term: str = '*:*',
               facet_fields: List[FIELD] = None,
               **params) -> SolrQueryResult:
        """
        Query a SOLR endpoint for a list of objects

        :param search_term:
        :param solr_params:
        :param target_class:
        :param params: key-value parameters. Keys should be in the schema
        :return:
        """
        sq = self.generate_query(target_class=target_class, **params)
        sq.other_params = solr_params
        sq.search_term = search_term
        sq.facet_fields = facet_fields
        rawres = self.execute(sq)
        items = [self.fetch_object(row, sq, target_class=target_class) for row in rawres.docs]
        return SolrQueryResult(items=items, response=rawres,
                               raw=rawres.raw_response,
                               start=rawres.raw_response['response']['start'],
                               num_found=rawres.hits)

    def generate_query(self, target_class: Type[YAMLRoot] = None, **params) -> SolrQuery:
        """
        Generate a solr query given query parameters

        :param prefixmap:
        :param params:
        :return:
        """
        sq = SolrQuery(prefixmap={})
        if target_class is not None:
            if self.discriminator_field is not None:
                params[self.discriminator_field] = target_class.class_name
        self._generate_query_for_params(sq, params)
        return sq

    def _generate_query_for_params(self, sq: SolrQuery, params: Dict) -> None:
        schema = self.schema
        mapper = self.mapper
        for sn, v in params.items():
            slot = mapper._get_slot(sn)
            if slot is not None:
                slot_range = slot.range
            else:
                slot_range = None
                logging.error(f'Unknown slot name: {sn}')
            solr_prop = mapper._slot_to_solr_prop(slot, sq.prefixmap)
            solr_val = mapper.pyval_to_solr_atom(v, range=slot_range, query=sq)
            sq.add_constraint(solr_prop, solr_val)

    def fetch_object(self, row: Dict,
                     original_query: SolrQuery = None,
                     target_class: Type[YAMLRoot] = None) -> YAMLRoot:
        """
        Given an ID, query out other fields and populate object
        :param row:
        :param original_query:
        :param target_class:
        :return:
        """
        mapper = self.mapper
        new_obj = {}
        schema_class = self.schema.classes[target_class.class_name]
        for k, v in row.items():
            if k == '_version_':
                # TODO
                continue
            if v is not None and v != []:
                slot = mapper._lookup_slot(schema_class, k)
                if slot is None:
                    if k == 'id':
                        # autogen field
                        continue
                    else:
                        raise ValueError(f'Cannot retrieve slot for field {k}')
                if not slot.multivalued and isinstance(v, list):
                    if len(v) == 1:
                        v = v[0]
                    elif len(v) == 0:
                        v = None
                    else:
                        raise Exception(f'Multi-values for scalar field {schema_class.name}.{k} == {v} {len(v)}')
                new_obj[k] = v
        cls = mapper._get_linkml_class(new_obj)
        if cls is None:
            cls = target_class
        logging.debug(new_obj)
        return cls(**new_obj)

    def fetch_by_id(self, q: AbstractQuery) -> YAMLRoot:
        if isinstance(q, FetchById):
            tgt_classes = [c for c in self.python_classes if c.class_name == q.target_class]
            id_field = None
            for c in self.schema.classes.values():
                if c.name == q.target_class:
                    for s in c.slots:
                        if self.schema.slots[s].identifier:
                            id_field = s
            if not id_field:
                raise ValueError(f'No ID found for {q.target_class}')
            params = {'target_class': tgt_classes[0],
                     id_field: q.id}
            result = self.search(**params)
            return result.items

    def simple_query(self, target_class: str, **kwargs) -> List[YAMLRoot]:
        tgt_classes = [c for c in self.python_classes if c.class_name == target_class]
        if len(tgt_classes) != 1:
            raise ValueError(f'Class: {target_class}')
        py_class = tgt_classes[0]
        params = {k: v for k, v in kwargs.items() if v is not None}
        result = self.search(target_class=py_class, **params)
        return result.items



    def execute(self, query: SolrQuery) -> RawSolrResult:
        """
        Execute a solr query on endpoint

        Endpoint can be an in-memory graph or remote endpoint

        :param query:
        :return:
        """
        #solr = pysolr.Solr(self.endpoint.url, **solr_params)
        solr = pysolr.Solr(self.endpoint.url)
        params = query.http_params()
        #logging.info(params)
        print(f'Params={params}')
        results = solr.search(query.search_term, **params)
        if solr.session is not None:
            solr.session.close()
        return results

    def add(self, objs: List[YAMLRoot], commit=True):
        """
        Adds an instance of a LinkML class as a Solr document

        :param objs: list of objects to add
        :return:
        """
        jstrs = [json_dumper.dumps(obj, inject_type=False) for obj in objs]
        nu_objs = [json.loads(s) for s in jstrs]
        print(f'Adding = {nu_objs}')
        solr = pysolr.Solr(self.endpoint.url)
        r = solr.add(nu_objs)
        if commit:
            solr.commit()
        return r

    def delete_all(self, commit=True):
        """
        Deletes all documents in core

        :param commit:
        :return:
        """
        solr = pysolr.Solr(self.endpoint.url)
        solr.delete(q='*:*')
        if commit:
            solr.commit()

    def _solr_request(self, req: Dict, path='schema'):
        response = requests.post(f'{self.endpoint.url}/{path}',
                                 headers={"Content-Type": "application/json"},
                                 data=json.dumps(req, indent=' '))
        if response.status_code != 200:
            logging.error(f'Failed to execute {path} {req}: {response.status_code} :: {response.text}')
        return response

    def _solr_query(self, path='schema', strict=True):
        response = requests.get(f'{self.endpoint.url}/{path}',
                                 headers={"Content-Type": "application/json"})
        if response.status_code != 200:
            # TODO: raise exception
            logging.error(f'Failed to execute {path}: {response.status_code} :: {response.text}')
        return response

    def _response_json(self, response, strict=True):
        try:
            return response.json()
        except JSONDecodeError as e:
            if strict:
                raise JSONDecodeError(e)
            else:
                return {}

    def _get_fields(self, strict=True):
        response = self._solr_query(path='schema/fields', strict=strict)
        obj = self._response_json(response, strict=strict)
        return [f['name'] for f in obj.get('fields', [])]

    def _get_fieldtypes(self, strict=True):
        response = self._solr_query(path='schema/fieldtypes', strict=strict)
        obj = self._response_json(response, strict=strict)
        return [f['name'] for f in obj.get('fieldTypes', [])]

    def load_schema(self):
        """
        Adds a schema to SOLR corresponding to a LinkML schema
        """
        existing_fields = self._get_fields(strict=False)
        existing_fieldtypes = self._get_fieldtypes(strict=False)
        logging.info(f'Current Fields={existing_fields}')
        if 'int' not in existing_fieldtypes:
            self._solr_request({"add-field-type": {
                'name': 'int',
                'class': 'solr.TrieIntField',
                'precisionStep': "0",
                "positionIncrementGap": "0"
            }}, path='schema')
        if 'date' not in existing_fieldtypes:
            self._solr_request({"add-field-type": {
                'name': 'date',
                'class': 'solr.TrieDateField',
            }}, path='schema')
        gen = SolrSchemaGenerator(self.schema)
        gen.serialize()
        post_obj = gen.post_request
        for f in post_obj['add-field']:
            if f['name'] not in existing_fields:
                self._solr_request({'add-field': f}, path='schema')




# Auto generated from solrschema.yaml by pythongen.py version: 0.9.0
# Generation date: 2021-07-25 18:10
# Schema: solr-schema
#
# id: https://w3id.org/linkml/solr
# description: TODO
# license: https://creativecommons.org/publicdomain/zero/1.0/

import dataclasses
import sys
import re
from jsonasobj2 import JsonObj, as_dict
from typing import Optional, List, Union, Dict, ClassVar, Any
from dataclasses import dataclass
from linkml_runtime.linkml_model.meta import EnumDefinition, PermissibleValue, PvFormulaOptions

from linkml_runtime.utils.slot import Slot
from linkml_runtime.utils.metamodelcore import empty_list, empty_dict, bnode
from linkml_runtime.utils.yamlutils import YAMLRoot, extended_str, extended_float, extended_int
from linkml_runtime.utils.dataclass_extensions_376 import dataclasses_init_fn_with_kwargs
from linkml_runtime.utils.formatutils import camelcase, underscore, sfx
from linkml_runtime.utils.enumerations import EnumDefinitionImpl
from rdflib import Namespace, URIRef
from linkml_runtime.utils.curienamespace import CurieNamespace
from linkml_runtime.linkml_model.types import Boolean, Integer, String
from linkml_runtime.utils.metamodelcore import Bool

metamodel_version = "1.7.0"

# Overwrite dataclasses _init_fn to add **kwargs in __init__
dataclasses._init_fn = dataclasses_init_fn_with_kwargs

# Namespaces
BOOLEAN = CurieNamespace('boolean', 'http://example.org/UNKNOWN/boolean/')
LINKML = CurieNamespace('linkml', 'https://w3id.org/linkml/')
SOLR = CurieNamespace('solr', 'https://w3id.org/linkml/solr/')
DEFAULT_ = SOLR


# Types

# Class references
class FieldName(extended_str):
    pass


class DynamicFieldName(extended_str):
    pass


class FieldTypeName(extended_str):
    pass


@dataclass
class Field(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.Field
    class_class_curie: ClassVar[str] = "solr:Field"
    class_name: ClassVar[str] = "Field"
    class_model_uri: ClassVar[URIRef] = SOLR.Field

    name: Union[str, FieldName] = None
    type: Optional[str] = None
    stored: Optional[str] = None
    indexed: Optional[str] = None
    multiValued: Optional[Union[bool, Bool]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, FieldName):
            self.name = FieldName(self.name)

        if self.type is not None and not isinstance(self.type, str):
            self.type = str(self.type)

        if self.stored is not None and not isinstance(self.stored, str):
            self.stored = str(self.stored)

        if self.indexed is not None and not isinstance(self.indexed, str):
            self.indexed = str(self.indexed)

        if self.multiValued is not None and not isinstance(self.multiValued, Bool):
            self.multiValued = Bool(self.multiValued)

        super().__post_init__(**kwargs)


@dataclass
class DynamicField(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.DynamicField
    class_class_curie: ClassVar[str] = "solr:DynamicField"
    class_name: ClassVar[str] = "DynamicField"
    class_model_uri: ClassVar[URIRef] = SOLR.DynamicField

    name: Union[str, DynamicFieldName] = None
    type: Optional[str] = None
    stored: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, DynamicFieldName):
            self.name = DynamicFieldName(self.name)

        if self.type is not None and not isinstance(self.type, str):
            self.type = str(self.type)

        if self.stored is not None and not isinstance(self.stored, str):
            self.stored = str(self.stored)

        super().__post_init__(**kwargs)


@dataclass
class CopyField(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.CopyField
    class_class_curie: ClassVar[str] = "solr:CopyField"
    class_name: ClassVar[str] = "CopyField"
    class_model_uri: ClassVar[URIRef] = SOLR.CopyField

    source: str = None
    dest: Union[str, List[str]] = None
    maxChars: Optional[int] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.source):
            self.MissingRequiredField("source")
        if not isinstance(self.source, str):
            self.source = str(self.source)

        if self._is_empty(self.dest):
            self.MissingRequiredField("dest")
        if not isinstance(self.dest, list):
            self.dest = [self.dest] if self.dest is not None else []
        self.dest = [v if isinstance(v, str) else str(v) for v in self.dest]

        if self.maxChars is not None and not isinstance(self.maxChars, int):
            self.maxChars = int(self.maxChars)

        super().__post_init__(**kwargs)


@dataclass
class FieldType(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.FieldType
    class_class_curie: ClassVar[str] = "solr:FieldType"
    class_name: ClassVar[str] = "FieldType"
    class_model_uri: ClassVar[URIRef] = SOLR.FieldType

    name: Union[str, FieldTypeName] = None
    _class: Optional[str] = None
    positionIncrementGap: Optional[str] = None
    analyzer: Optional[Union[dict, "Analyzer"]] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._is_empty(self.name):
            self.MissingRequiredField("name")
        if not isinstance(self.name, FieldTypeName):
            self.name = FieldTypeName(self.name)

        if self._class is not None and not isinstance(self._class, str):
            self._class = str(self._class)

        if self.positionIncrementGap is not None and not isinstance(self.positionIncrementGap, str):
            self.positionIncrementGap = str(self.positionIncrementGap)

        if self.analyzer is not None and not isinstance(self.analyzer, Analyzer):
            self.analyzer = Analyzer(**as_dict(self.analyzer))

        super().__post_init__(**kwargs)


@dataclass
class Analyzer(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.Analyzer
    class_class_curie: ClassVar[str] = "solr:Analyzer"
    class_name: ClassVar[str] = "Analyzer"
    class_model_uri: ClassVar[URIRef] = SOLR.Analyzer

    charFilters: Optional[Union[str, List[str]]] = empty_list()
    tokenizer: Optional[Union[dict, "Tokenizer"]] = None
    filters: Optional[Union[str, List[str]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if not isinstance(self.charFilters, list):
            self.charFilters = [self.charFilters] if self.charFilters is not None else []
        self.charFilters = [v if isinstance(v, str) else str(v) for v in self.charFilters]

        if self.tokenizer is not None and not isinstance(self.tokenizer, Tokenizer):
            self.tokenizer = Tokenizer(**as_dict(self.tokenizer))

        if not isinstance(self.filters, list):
            self.filters = [self.filters] if self.filters is not None else []
        self.filters = [v if isinstance(v, str) else str(v) for v in self.filters]

        super().__post_init__(**kwargs)


@dataclass
class Tokenizer(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.Tokenizer
    class_class_curie: ClassVar[str] = "solr:Tokenizer"
    class_name: ClassVar[str] = "Tokenizer"
    class_model_uri: ClassVar[URIRef] = SOLR.Tokenizer

    _class: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._class is not None and not isinstance(self._class, str):
            self._class = str(self._class)

        super().__post_init__(**kwargs)


@dataclass
class CharFilter(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.CharFilter
    class_class_curie: ClassVar[str] = "solr:CharFilter"
    class_name: ClassVar[str] = "CharFilter"
    class_model_uri: ClassVar[URIRef] = SOLR.CharFilter

    _class: Optional[str] = None
    replacement: Optional[str] = None
    pattern: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._class is not None and not isinstance(self._class, str):
            self._class = str(self._class)

        if self.replacement is not None and not isinstance(self.replacement, str):
            self.replacement = str(self.replacement)

        if self.pattern is not None and not isinstance(self.pattern, str):
            self.pattern = str(self.pattern)

        super().__post_init__(**kwargs)


@dataclass
class Filter(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.Filter
    class_class_curie: ClassVar[str] = "solr:Filter"
    class_name: ClassVar[str] = "Filter"
    class_model_uri: ClassVar[URIRef] = SOLR.Filter

    _class: Optional[str] = None
    preserveOriginal: Optional[str] = None

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        if self._class is not None and not isinstance(self._class, str):
            self._class = str(self._class)

        if self.preserveOriginal is not None and not isinstance(self.preserveOriginal, str):
            self.preserveOriginal = str(self.preserveOriginal)

        super().__post_init__(**kwargs)


@dataclass
class Transaction(YAMLRoot):
    _inherited_slots: ClassVar[List[str]] = []

    class_class_uri: ClassVar[URIRef] = SOLR.Transaction
    class_class_curie: ClassVar[str] = "solr:Transaction"
    class_name: ClassVar[str] = "transaction"
    class_model_uri: ClassVar[URIRef] = SOLR.Transaction

    add_field: Optional[Union[Dict[Union[str, FieldName], Union[dict, Field]], List[Union[dict, Field]]]] = empty_dict()
    delete_field: Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]] = empty_list()
    replace_field: Optional[Union[Union[str, FieldName], List[Union[str, FieldName]]]] = empty_list()
    add_field_type: Optional[Union[Dict[Union[str, FieldTypeName], Union[dict, FieldType]], List[Union[dict, FieldType]]]] = empty_dict()
    delete_field_type: Optional[Union[Union[str, FieldTypeName], List[Union[str, FieldTypeName]]]] = empty_list()
    replace_field_type: Optional[Union[Union[str, FieldTypeName], List[Union[str, FieldTypeName]]]] = empty_list()
    add_dynamic_field: Optional[Union[Dict[Union[str, DynamicFieldName], Union[dict, DynamicField]], List[Union[dict, DynamicField]]]] = empty_dict()
    delete_dynamic_field: Optional[Union[Union[str, DynamicFieldName], List[Union[str, DynamicFieldName]]]] = empty_list()
    replace_dynamic_field: Optional[Union[Union[str, DynamicFieldName], List[Union[str, DynamicFieldName]]]] = empty_list()

    def __post_init__(self, *_: List[str], **kwargs: Dict[str, Any]):
        self._normalize_inlined_as_list(slot_name="add_field", slot_type=Field, key_name="name", keyed=True)

        if not isinstance(self.delete_field, list):
            self.delete_field = [self.delete_field] if self.delete_field is not None else []
        self.delete_field = [v if isinstance(v, FieldName) else FieldName(v) for v in self.delete_field]

        if not isinstance(self.replace_field, list):
            self.replace_field = [self.replace_field] if self.replace_field is not None else []
        self.replace_field = [v if isinstance(v, FieldName) else FieldName(v) for v in self.replace_field]

        self._normalize_inlined_as_list(slot_name="add_field_type", slot_type=FieldType, key_name="name", keyed=True)

        if not isinstance(self.delete_field_type, list):
            self.delete_field_type = [self.delete_field_type] if self.delete_field_type is not None else []
        self.delete_field_type = [v if isinstance(v, FieldTypeName) else FieldTypeName(v) for v in self.delete_field_type]

        if not isinstance(self.replace_field_type, list):
            self.replace_field_type = [self.replace_field_type] if self.replace_field_type is not None else []
        self.replace_field_type = [v if isinstance(v, FieldTypeName) else FieldTypeName(v) for v in self.replace_field_type]

        self._normalize_inlined_as_list(slot_name="add_dynamic_field", slot_type=DynamicField, key_name="name", keyed=True)

        if not isinstance(self.delete_dynamic_field, list):
            self.delete_dynamic_field = [self.delete_dynamic_field] if self.delete_dynamic_field is not None else []
        self.delete_dynamic_field = [v if isinstance(v, DynamicFieldName) else DynamicFieldName(v) for v in self.delete_dynamic_field]

        if not isinstance(self.replace_dynamic_field, list):
            self.replace_dynamic_field = [self.replace_dynamic_field] if self.replace_dynamic_field is not None else []
        self.replace_dynamic_field = [v if isinstance(v, DynamicFieldName) else DynamicFieldName(v) for v in self.replace_dynamic_field]

        super().__post_init__(**kwargs)


# Enumerations


# Slots




from typing import List
import click
import time
import logging
import subprocess

from linkml_runtime.linkml_model import SchemaDefinition
from linkml_runtime.loaders import yaml_loader
from linkml_solr import SolrQueryEngine, SolrEndpoint, DEFAULT_CORE, DEFAULT_SOLR_URL
from linkml_solr.utils.solr_bulkload import bulkload_file


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
def main(verbose: int, quiet: bool):
    """LinkML-Solr utility commands

    Args:

        verbose (int): Verbose.
        quiet (bool): Quiet.

    Returns:

        None.

    """
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)
    if quiet:
        logging.basicConfig(level=logging.ERROR)

@main.command()
@click.option('--schema', '-s',
              help='Path to schema.')
@click.option('--core', '-C',
              default=DEFAULT_CORE,
              show_default=True,
              help='solr core.')
@click.option('--format', '-f',
              default='csv',
              show_default=True,
              help='input format.')
@click.option('--url', '-u',
              default=DEFAULT_SOLR_URL,
              help='solr url.')
@click.argument('files', nargs=-1)
def bulkload(files, format, schema, url, core):
    """
    Convert multiple golr yaml schemas to linkml
    """
    inputs = {}
    if schema is not None:
        with open(schema) as stream:
            schema_obj = yaml_loader.load(stream, target_class=SchemaDefinition)
    for f in files:
        bulkload_file(f, format=format, schema=schema_obj, core=core, base_url=url)

@main.command()
@click.option('--schema', '-s',
              help='Path to LinkML yaml schema.')
@click.option('--core', '-C',
              default=DEFAULT_CORE,
              show_default=True,
              help='name of solr core.')
@click.option('--container', '-n',
              default='my_solr',
              show_default=True,
              help='name of docker container')
@click.option('--port', '-P',
              default='8983',
              show_default=True,
              help='http port number on which solr rules.')
@click.option('--sleep',
              default=10,
              show_default=True,
              help='Number of seconds to sleep after initiating server start before exiting.')
@click.option('--create-schema/--no-create-schema',
              default=True,
              show_default=True,
              help='create the solr schema from the linkml schema on startup')
@click.option('--kill/--no-kill',
              default=True,
              show_default=True,
              help='kill and remove any existing container')
@click.option('--url', '-u',
              default=DEFAULT_SOLR_URL,
              show_default=True,
              help='solr url.')
def start_server(schema, kill, container, url, core, port, sleep: int, create_schema):
    """
    Starts a solr server (via Docker)
    """
    if kill:
        subprocess.run(['docker', 'kill', container])
        subprocess.run(['docker', 'rm', container])
    command = [
        'docker',
        'run',
        '--name',
        container,
        '-p',
        f'{port}:{port}',
        'solr:8',
        'solr-precreate',
        core]
    subprocess.Popen(command)
    print(f'Sleeping for {sleep} seconds to allow server time to initialize...')
    time.sleep(sleep)
    print(f'Done sleeping!')
    if create_schema and schema:
        # TODO: use SchemaView
        from linkml.generators.yamlgen import YAMLGenerator
        schema_obj = YAMLGenerator(schema).schema
        qe = SolrQueryEngine(schema=schema_obj,
                             endpoint=SolrEndpoint(url=f'{url}/{core}'))
        qe.load_schema()

@main.command()
@click.option('--schema', '-s',
              help='Path to LinkML yaml schema.')
@click.option('--container', '-n',
              default='my_solr',
              show_default=True,
              help='name of docker container')
@click.option('--port', '-P',
              default='8983',
              show_default=True,
              help='http port number on which solr rules.')
@click.option('--url', '-u',
              default=DEFAULT_SOLR_URL,
              show_default=True,
              help='solr url.')
@click.argument('cores', nargs=-1)
def add_cores(cores, schema, container, url, port):
    """
    Starts a solr server (via Docker)
    """
    for core in cores:
        command = [
            'docker',
            'exec',
            container,
            'solr',
            'create',
            '-c',
            core]
        subprocess.Popen(command)

@main.command()
@click.option('--schema', '-s',
              help='Path to schema.')
@click.option('--core', '-C',
              default=DEFAULT_CORE,
              show_default=True,
              help='solr core.')
@click.option('--url', '-u',
              default=DEFAULT_SOLR_URL,
              show_default=True,
              help='solr url.')
def create_schema(schema, url, core):
    """
    Creates a solr schema from a LinkML schema
    """
    # TODO: use schemaview
    from linkml.generators.yamlgen import YAMLGenerator
    schema_obj = YAMLGenerator(schema).schema
    qe = SolrQueryEngine(schema=schema_obj,
                         endpoint=SolrEndpoint(url=f'{url}/{core}'))
    qe.load_schema()


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['linkml_solr', 'linkml_solr.utils']

package_data = \
{'': ['*']}

install_requires = \
['linkml-dataops>=0.1.0,<0.2.0',
 'linkml-runtime>=1.2.7,<2.0.0',
 'linkml>=1.2.0,<2.0.0',
 'pysolr>=3.9.0,<4.0.0']

entry_points = \
{'console_scripts': ['lsolr = linkml_solr.cli:main']}

setup_kwargs = {
    'name': 'linkml-solr',
    'version': '0.1.3',
    'description': '',
    'long_description': None,
    'author': 'Chris Mungall',
    'author_email': 'cjm@berkeleybop.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.4,<4.0.0',
}


setup(**setup_kwargs)




from dataclasses import dataclass
from typing import Dict, List, Any, Optional

from linkml_runtime.linkml_model.meta import YAMLRoot

PREFIXMAP = Dict[str, str]
FIELD = str
RawSolrResult = Dict


@dataclass
class SolrEndpoint(object):
    url: Optional[str] = None
    type_property: str = 'type'


@dataclass
class SolrQuery:
    prefixmap: PREFIXMAP = None
    fields: List[FIELD] = None
    facet_fields: List[FIELD] = None
    filter_query: Dict[FIELD, Any] = None
    other_params: Dict[str, Any] = None
    search_term: str = '*:*'
    rows: int = 100

    def http_params(self) -> dict:
        params = {'rows': self.rows}
        if self.other_params is not None:
            for k, v in self.other_params.items():
                params[k] = v
        if self.filter_query is None:
            self.filter_query = {}
        fq = []
        for k, v in self.filter_query.items():
            if isinstance(v, list):
                for v1 in v:
                    fq.append(f'{k}:{_quote(v1)}')
            else:
                fq.append(f'{k}:{_quote(v)}')
        params['fq'] = fq
        if self.facet_fields is not None:
            params['facet'] = 'on'
            params['facet.field'] = self.facet_fields
            params['facet.limit'] = 25
            params['facet.mincount'] = 1
        if self.fields is not None:
            params['fields'] = ','.join(self.fields)
        return params

    def add_constraint(self, solr_prop, solr_val):
        if self.filter_query is None:
            self.filter_query = {}
        self.filter_query[solr_prop] = solr_val



@dataclass
class SolrQueryResult:
    num_found: int = 0
    start: int = 0
    items: List[YAMLRoot] = None
    facet_counts: Dict = None
    highlighting: str = None
    raw: Any = None
    response: Any = None

    def __post_init__(self):
        if self.response.facets is not None:
            ffs = self.response.facets.get('facet_fields', {})
            facet_counts = {}
            for slot, counts in ffs.items():
                facet_counts[slot] = {term: count for term, count in zip(*[iter(counts)]*2)}
            self.facet_counts = facet_counts


def _quote(v, operator="OR"):
    if isinstance(v, list):
        if len(v) == 1:
            return _quote(v[0], operator)
        else:
            v2 = f" {operator} ".join([_quote(x) for x in v])
            return f'({v2})'
    else:
        return f'"{v}"'



import click
from typing import List, Dict
import logging


import yaml

def golr_to_linkml_schema(inputs: Dict[str, dict], base_url: str = 'https://w3id.org/', schema_id: str = 'example') -> dict:
    cls_to_file = {}
    slots = {'document_category': {}}
    classes = {}
    url = f'{base_url}/{schema_id}'
    schema = {
        'id': url,
        'description': 'TODO',
        'imports': [
            'linkml:types'
        ],
        'prefixes': {
            'linkml': 'https://w3id.org/linkml/',
            schema_id: 'http://kbase.us/'
        },
        'default_prefix': schema_id,
        'classes': classes,
        'slots': slots,
        'types': {
            'SearchableString': {
                'typeof': 'string'
            }
        }
    }
    for fn, input in inputs.items():
        cslots = ['document_category']
        n = input['document_category']
        c = {'aliases': [input['display_name']],
            'description': input['description'],
            'slots': cslots}
        if n in classes:
            raise Exception(f'Duplicate class name: {n}: in {cls_to_file[n]} and {fn}')
        cls_to_file[n] = fn
        classes[n] = c
        for f in input['fields']:
            fid = f['id']

            card = f.get('cardinality', '')
            if card == 'multi':
                multivalued = True
            else:
                multivalued = False
            range = f['type']
            searchable = f.get('searchable', False)
            if searchable:
                range = 'SearchableString'
            slot = {'description': f['description'],
                    'range': range,
                    'aliases': [f['display_name']],
                    'multivalued': multivalued,
                    'comments': []}
            #for p in f.get('property', []):
            #    slot['aliases'].append(p) # TODO: xref
            if f.get('indexed', False):
                slot['comments'].append('Indexed')
            if fid in slot:
                logging.warning(f'Overwriting: {fid} in {n}')
            slots[fid] = slot
            cslots.append(fid)
            if searchable:
                sslot = slot.copy()
                sfid = f'{fid}_searchable'
                slots[sfid] = sslot
                cslots.append(sfid)
    return schema

@click.command()
@click.option('--schema', '-s',
              help='Name of schema.')
@click.argument('files', nargs=-1)
def convert(files, schema):
    """
    Convert multiple golr yaml schemas to linkml

    Note this is primarily useful for converting schemas in the
    golr format used by AmiGO/Monarch

    :param files:
    :param schema:
    :return:
    """
    inputs = {}
    for f in files:
        with open(f) as stream:
            inputs[f] = yaml.safe_load(stream)
    s = golr_to_linkml_schema(inputs, schema_id=schema)
    print(yaml.dump(s, sort_keys=False))

if __name__ == '__main__':
    convert()